/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  arrayaccess.c
 *
 *      Access within an array of 32-bit words
 *
 *           INT32       getDataBit()
 *           void        setDataBit()
 *           void        clearDataBit()
 *           INT32       getDataDibit()
 *           void        setDataDibit()
 *           void        clearDataDibit()
 *           INT32       getDataQbit()
 *           void        setDataQbit()
 *           void        clearDataQbit()
 *           INT32       getDataByte()
 *           void        setDataByte()
 *           INT32       getDataTwoBytes()
 *           void        setDataTwoBytes()
 *
 */


#include <stdio.h>

#include "allheaders.h"



/*----------------------------------------------------------------------*
 *                 Access within an array of 32-bit words               *
 *----------------------------------------------------------------------*/
/*!
 *  getDataBit()
 *
 *       Input: line  (ptr to beginning of data line)
 *              n     (pixel index)
 *       Return: val of the nth (1-bit) pixel.
 */
INT32
getDataBit(UINT32  *line,
	   INT32    n)
{
    return (*(line + (n >> 5)) >> (31 - (n & 31))) & 1;
}


/*!
 *  setDataBit()
 *
 *       Input: line  (ptr to beginning of data line)
 *              n     (pixel index)
 *       Return: void
 *       Action: sets the pixel to 1
 */
void
setDataBit(UINT32  *line,
	   INT32    n)
{
    *(line + (n >> 5)) |= (0x80000000 >> (n & 31));
}


/*!
 *  clearDataBit()
 *
 *       Input: line  (ptr to beginning of data line)
 *              n     (pixel index)
 *       Return: void
 *       Action: sets the (1-bit) pixel to 0
 */
void
clearDataBit(UINT32  *line,
	     INT32    n)
{
    *(line + (n >> 5)) &= ~(0x80000000 >> (n & 31));
}


/*!
 *  getDataDibit()
 *
 *       Input: line  (ptr to beginning of data line)
 *              n     (pixel index)
 *       Return: val of the nth (2-bit) pixel.
 */
INT32
getDataDibit(UINT32  *line,
	     INT32    n)
{
    return (*(line + (n >> 4)) >> (2 * (15 - (n & 15)))) & 3;
}


/*!
 *  setDataDibit()
 *
 *       Input: line  (ptr to beginning of data line)
 *              n     (pixel index)
 *              val   (val to be inserted: 0 - 3)
 *       Return: void
 */
void
setDataDibit(UINT32  *line,
	     INT32    n,
	     INT32    val)
{
UINT32    *pword;

    pword = line + (n >> 4);
    *pword &= ~(0xc0000000 >> (2 * (n & 15)));  /* clear */
    *pword |= val << (30 - 2 * (n & 15));   /* set */
    return;
}


/*!
 *  clearDataDibit()
 *
 *       Input: line  (ptr to beginning of data line)
 *              n     (pixel index)
 *       Return: void
 *       Action: sets the (2-bit) pixel to 0
 */
void
clearDataDibit(UINT32  *line,
	       INT32    n)
{
    *(line + (n >> 4)) &= ~(0xc0000000 >> (2 * (n & 15)));
}


/*!
 *  getDataQbit()
 *
 *       Input: line  (ptr to beginning of data line)
 *              n     (pixel index)
 *       Return: val of the nth (4-bit) pixel.
 */
INT32
getDataQbit(UINT32  *line,
	    INT32    n)
{
    return (*(line + (n >> 3)) >> (4 * (7 - (n & 7)))) & 0xf;
}


/*!
 *  setDataQbit()
 *
 *       Input: line  (ptr to beginning of data line)
 *              n     (pixel index)
 *              val   (val to be inserted: 0 - 0xf)
 *       Return: void
 */
void
setDataQbit(UINT32  *line,
	    INT32    n,
	    INT32    val)
{
UINT32    *pword;

    pword = line + (n >> 3);
    *pword &= ~(0xf0000000 >> (4 * (n & 7)));  /* clear */
    *pword |= val << (28 - 4 * (n & 7));   /* set */
    return;
}


/*!
 *  clearDataQbit()
 *
 *       Input: line  (ptr to beginning of data line)
 *              n     (pixel index)
 *       Return: void
 *       Action: sets the (4-bit) pixel to 0
 */
void
clearDataQbit(UINT32  *line,
	      INT32    n)
{
    *(line + (n >> 3)) &= ~(0xf0000000 >> (4 * (n & 7)));
}


/*!
 *  getDataByte()
 *
 *       Input: line  (ptr to beginning of data line)
 *              n     (pixel index)
 *       Return: value of the n-th (byte) pixel
 */
INT32
getDataByte(UINT32  *line,
	    INT32    n)
{
#ifdef  _BIG_ENDIAN_
    return *((UINT8 *)line + n);
#else  /* _LITTLE_ENDIAN_ */
    return *(UINT8 *)((UINT32)((UINT8 *)line + n) ^ 3);
#endif  /* _BIG_ENDIAN_ */
}


/*!
 *  setDataByte()
 *
 *       Input: line  (ptr to beginning of data line)
 *              n     (pixel index)
 *              val   (val to be inserted: 0 - 0xff)
 *       Return: void
 */
void
setDataByte(UINT32  *line,
	    INT32    n,
	    INT32    val)
{
#ifdef  _BIG_ENDIAN_
    *((UINT8 *)line + n) = val;
#else  /* _LITTLE_ENDIAN_ */
    *(UINT8 *)((UINT32)((UINT8 *)line + n) ^ 3) = val;
#endif  /* _BIG_ENDIAN_ */
}



/*!
 *  getDataTwoBytes()
 *
 *       Input: line  (ptr to beginning of data line)
 *              n     (pixel index)
 *       Return: value of the n-th (2-byte) pixel
 */
INT32
getDataTwoBytes(UINT32  *line,
	        INT32    n)
{
#ifdef  _BIG_ENDIAN_
    return *((UINT16 *)line + n);
#else  /* _LITTLE_ENDIAN_ */
    return *(UINT16 *)((UINT32)((UINT16 *)line + n) ^ 2);
#endif  /* _BIG_ENDIAN_ */
}


/*!
 *  setDataTwoBytes()
 *
 *       Input: line  (ptr to beginning of data line)
 *              n     (pixel index)
 *              val   (val to be inserted: 0 - 0xffff)
 *       Return: void
 */
void
setDataTwoBytes(UINT32  *line,
	        INT32    n,
	        INT32    val)
{
#ifdef  _BIG_ENDIAN_
    *((UINT16 *)line + n) = val;
#else  /* _LITTLE_ENDIAN_ */
    *(UINT16 *)((UINT32)((UINT16 *)line + n) ^ 2) = val;
#endif  /* _BIG_ENDIAN_ */
}


