/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *   pixa.c
 *
 *      Pixa creation, destruction, copying
 *           PIXA   *pixaCreate()
 *           PIXA   *pixaCreateFromPix()
 *           void    pixaDestroy()
 *           PIXA   *pixaCopy()
 *
 *      Pixa addition
 *           INT32   pixaAddPix()
 *           INT32   pixaExtendArray()
 *           INT32   pixaAddBox()
 *
 *      Pixa accessors
 *           INT32   pixaGetCount()
 *           INT32   pixaChangeRefcount()
 *           PIX    *pixaGetPix()
 *           BOX    *pixaGetBox()
 *
 *      Pixaa creation, destruction
 *           PIXAA  *pixaaCreate()
 *           void    pixaaDestroy()
 *
 *      Pixaa addition
 *           INT32   pixaaAddPixa()
 *           INT32   pixaaExtendArray()
 *
 *      Pixaa accessors
 *           INT32   pixaaGetCount()
 *           PIXA   *pixaaGetPixa()
 *
 *      Pixa and Pixaa Display
 *           PIX    *pixaDisplay()
 *           PIX    *pixaDisplayOnLattice()
 *           PIX    *pixaaDisplay()
 *           PIX    *pixaaDisplayByPixa()
 *
 *      Pixa serialized I/O
 *           INT32   pixaWriteStream()
 *           PIXA   *pixaReadStream()
 *
 *   
 *   Note: The reference counting for the pixa is analogous
 *         to that for the pix.  Every pixa has a ref count
 *         of 1 when it is created.  The only way this ref
 *         count is incremented is by accessing a pixa in a
 *         pixaa using pixaaGetPixa() with the NEWREF flag.
 *         Using either the COPY or CLONE flag makes a new
 *         pixa with a ref count of 1; use of the CLONE flag
 *         clones rather than copies each pix in the array.
 *         The programmer only needs to remember that
 *         regardless how a pixa is obtained (e.g., from
 *         pixaCreate*(), pixaCopy(), or a pixaa accessor),
 *         it is necessary to call pixaDestroy() on it.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>   /* for sqrt() */

#include "allheaders.h"


#define  INITIAL_PTR_ARRAYSIZE      20    /* n'import quoi */




/*---------------------------------------------------------------------*
 *                    Pixa creation, destruction, copy                 *
 *---------------------------------------------------------------------*/
/*!
 *  pixaCreate()
 *       Input:  n  (initial number of ptrs)
 *       Return: pixa, or null on error
 */
PIXA *
pixaCreate(INT32  n)
{
PIXA   *pixa;

    PROCNAME("pixaCreate");

    if (n <= 0)
	n = INITIAL_PTR_ARRAYSIZE;

    if ((pixa = (PIXA *)CALLOC(1, sizeof(PIXA))) == NULL)
	return (PIXA *)ERROR_PTR("pixa not made", procName, NULL);
    pixa->n = 0;
    pixa->nalloc = n;
    pixa->refcount = 1;
    
    if ((pixa->pix = (PIX **)CALLOC(n, sizeof(PIX *))) == NULL)
	return (PIXA *)ERROR_PTR("pix ptrs not made", procName, NULL);

    if ((pixa->boxa = boxaCreate(n)) == NULL)
	return (PIXA *)ERROR_PTR("boxa not made", procName, NULL);

    return pixa;
}


/*!
 *  pixaCreateFromPix()
 *       Input:  pix  (with individual components on a lattice)
 *               n   (number of components)
 *               cellw   (width of each cell)
 *               cellh   (height of each cell)
 *       Return: pixa, or null on error
 *
 *       Note: we truncate each retrieved pix to the ON pixels,
 *             which we assume for now start at (0,0)
 */
PIXA *
pixaCreateFromPix(PIX   *pixs,
                  INT32  n,
                  INT32  cellw,
                  INT32  cellh)
{
INT32   w, h, nw, nh, i, j, index;
PIX    *pix, *pixt;
PIXA   *pixa;

    PROCNAME("pixaCreateFromPix");

    if (!pixs)
	return (PIXA *)ERROR_PTR("pixs not defined", procName, NULL);
    if (n <= 0)
	return (PIXA *)ERROR_PTR("n must be > 0", procName, NULL);

    if ((pixa = pixaCreate(n)) == NULL)
	return (PIXA *)ERROR_PTR("pixa not made", procName, NULL);
    if ((pixt = pixCreate(cellw, cellh, 1)) == NULL)
	return (PIXA *)ERROR_PTR("pixt not made", procName, NULL);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    nw = w / cellw;
    nh = h / cellh;
    for (i = 0, index = 0; i < nh; i++) {
	for (j = 0; j < nw && index < n; j++, index++) {
	    pixRasterop(pixt, 0, 0, cellw, cellh, PIX_SRC, pixs,
	           j * cellw, i * cellh);
	    if (pixClipToForeground(pixt, &pix, NULL))
	        pixaAddPix(pixa, pix, INSERT);
	}
    }

    pixDestroy(&pixt);
    return pixa;
}


/*!
 *  pixaDestroy()
 *
 *       Input:  &pixa (<can be nulled>)
 *       Return:  void
 */
void
pixaDestroy(PIXA  **ppixa)
{
INT32   i;
PIXA   *pixa;

    PROCNAME("pixaDestroy");

    if (ppixa == NULL) {
        WARNING("ptr address is NULL!", procName);
	return;
    }

    if ((pixa = *ppixa) == NULL)
	return;

	/* if there are other refs, just decrement counter */
    pixaChangeRefcount(pixa, -1);
    if (pixa->refcount > 0) {
	*ppixa = NULL;
	return;
    }

    for (i = 0; i < pixa->n; i++)
	pixDestroy(&pixa->pix[i]);
    FREE((void *)pixa->pix);
    boxaDestroy(&pixa->boxa);

    FREE((void *)pixa);
    *ppixa = NULL;

    return;
}


/*!
 *  pixaCopy()
 *
 *     Input:  pixas
 *             copyflag (COPY to copy each pix and each box;
 *                       CLONE to use clones)
 *     Return: new pixa, with a copy or a clone of each pix
 */
PIXA *
pixaCopy(PIXA   *pixas,
         INT32   copyflag)
{
INT32   i;
BOX    *boxc;
BOXA   *boxa;
PIX    *pixc;
PIXA   *pixac;

    PROCNAME("pixaCopy");

    if (!pixas)
	return (PIXA *)ERROR_PTR("pixas not defined", procName, NULL);
    if (copyflag != COPY && copyflag != CLONE)
	return (PIXA *)ERROR_PTR("copyflag must be in {1,2}", procName, NULL);

    if ((pixac = pixaCreate(pixas->n)) == NULL)
	return (PIXA *)ERROR_PTR("pixac not made", procName, NULL);

    for (i = 0; i < pixas->n; i++) {
	if ((pixc = pixaGetPix(pixas, i, copyflag)) == NULL)
	    return (PIXA *)ERROR_PTR("pixc not made", procName, NULL);
	pixaAddPix(pixac, pixc, INSERT);
    }

    boxa = pixas->boxa;
    for (i = 0; i < boxa->n; i++) {
	if ((boxc = boxaGetBox(boxa, i, copyflag)) == NULL)
	    return (PIXA *)ERROR_PTR("boxc not made", procName, NULL);
	boxaAddBox(pixac->boxa, boxc, INSERT);
    }

    return pixac;
}



/*---------------------------------------------------------------------*
 *                              Pixa addition                          *
 *---------------------------------------------------------------------*/
/*!
 *  pixaAddPix()
 *
 *       Input:  pixa
 *               pix  (to be added)
 *               copyflag (INSERT, COPY, CLONE)
 */
INT32
pixaAddPix(PIXA   *pixa,
           PIX    *pix,
           INT32   copyflag)
{
INT32   n;
PIX    *pixc;

    PROCNAME("pixaAddPix");

    if (!pixa)
	return ERROR_INT("pixa not defined", procName, 1);
    if (!pix)
	return ERROR_INT("pix not defined", procName, 1);
    if (copyflag != INSERT && copyflag != COPY && copyflag != CLONE)
	return ERROR_INT("copyflag must be in {1,2,3}", procName, 1);

    if (copyflag == INSERT)
        pixc = pix;
    else if (copyflag == COPY) {
	if ((pixc = pixCopy(NULL, pix)) == NULL)
	    return ERROR_INT("pixc not made", procName, 1);
    }
    else  /* CLONE */
	pixc = pixClone(pix);

    n = pixaGetCount(pixa);
    if (n >= pixa->nalloc)
	pixaExtendArray(pixa);
    pixa->pix[n] = pixc;
    pixa->n++;

    return 0;
}


/*!
 *  pixaExtendArray()
 *
 *       Input:  pixa
 *       Return: 0 if OK; 1 on error
 */
INT32
pixaExtendArray(PIXA   *pixa)
{

    PROCNAME("pixaExtendArray");

    if (!pixa)
	return ERROR_INT("pixa not defined", procName, 1);

    if ((pixa->pix = (PIX **)reallocNew((void **)&pixa->pix,
                                4 * pixa->nalloc,
				8 * pixa->nalloc)) == NULL)
	return ERROR_INT("new ptr array not returned", procName, 1);

    pixa->nalloc = 2 * pixa->nalloc;
    return 0;
}


/*!
 *  pixaAddBox()
 *
 *       Input:  pixa
 *               box
 *               copyflag (INSERT, COPY, CLONE)
 *       Return: 0 if OK, 1 on error
 */
INT32
pixaAddBox(PIXA   *pixa,
           BOX    *box,
	   INT32   copyflag)
{
    PROCNAME("pixaAddBox");

    if (!pixa)
	return ERROR_INT("pixa not defined", procName, 1);
    if (!box)
	return ERROR_INT("box not defined", procName, 1);
    if (copyflag != INSERT && copyflag != COPY && copyflag != CLONE)
	return ERROR_INT("copyflag must be in {0,1,2}", procName, 1);
    if (!pixa->boxa)
	return ERROR_INT("boxa not defined", procName, 1);

    boxaAddBox(pixa->boxa, box, copyflag);
    return 0;
}



/*---------------------------------------------------------------------*
 *                             Pixa accessors                          *
 *---------------------------------------------------------------------*/
/*!
 *  pixaGetCount()
 *
 *       Input:  pixa
 *       Return: count, or 0 if no pixa
 */
INT32
pixaGetCount(PIXA   *pixa)
{

    PROCNAME("pixaGetCount");

    if (!pixa)
	return ERROR_INT("pixa not defined", procName, 0);

    return pixa->n;
}


/*!
 *  pixaChangeRefcount()
 *
 *       Input:  pixa
 *       Return: 0 if OK, 1 on error
 */
INT32
pixaChangeRefcount(PIXA   *pixa,
                   INT32   delta)
{
    PROCNAME("pixaChangeRefcount");

    if (!pixa)
	return ERROR_INT("pixa not defined", procName, 1);

    pixa->refcount += delta;
    return 0;
}


/*!
 *  pixaGetPix()
 *
 *       Input:  pixa
 *               index  (to the index-th pix)
 *               accesstype  (COPY or CLONE)
 *       Return: pix, or null on error
 */
PIX *
pixaGetPix(PIXA   *pixa,
           INT32   index,
	   INT32   accesstype)
{
    PROCNAME("pixaGetPix");

    if (!pixa)
	return (PIX *)ERROR_PTR("pixa not defined", procName, NULL);
    if (index < 0 || index >= pixa->n)
	return (PIX *)ERROR_PTR("index not valid", procName, NULL);
    if (accesstype != COPY && accesstype != CLONE)
	return (PIX *)ERROR_PTR("accesstype must be in {1,2}", procName, NULL);

    if (accesstype == COPY)
	return pixCopy(NULL, pixa->pix[index]);
    else  /* CLONE */
	return pixClone(pixa->pix[index]);
}


/*!
 *  pixaGetBox()
 *
 *       Input:  pixa
 *               index  (to the index-th pix)
 *               accesstype  (COPY or CLONE)
 *       Return: box, or null on error
 */
BOX *
pixaGetBox(PIXA   *pixa,
           INT32   index,
	   INT32   accesstype)
{
    PROCNAME("pixaGetBox");

    if (!pixa)
	return (BOX *)ERROR_PTR("pixa not defined", procName, NULL);
    if (accesstype != COPY && accesstype != CLONE)
	return (BOX *)ERROR_PTR("accesstype must be in {1,2}", procName, NULL);
    if (!pixa->boxa)
	return (BOX *)ERROR_PTR("boxa not defined", procName, NULL);
    if (index < 0 || index >= pixa->boxa->n)
	return (BOX *)ERROR_PTR("index not valid", procName, NULL);

    if (accesstype == COPY)
	return boxCopy(pixa->boxa->box[index]);
    else  /* CLONE */
	return boxClone(pixa->boxa->box[index]);
}




/*---------------------------------------------------------------------*
 *                    Pixaa creation and destruction                   *
 *---------------------------------------------------------------------*/
/*!
 *  pixaaCreate()
 *       Input:  n  (initial number of ptrs)
 *       Return: pixaa, or null on error
 */
PIXAA *
pixaaCreate(INT32  n)
{
PIXAA   *pixaa;

    PROCNAME("pixaaCreate");

    if (n <= 0)
	n = INITIAL_PTR_ARRAYSIZE;

    if ((pixaa = (PIXAA *)CALLOC(1, sizeof(PIXAA))) == NULL)
	return (PIXAA *)ERROR_PTR("pixaa not made", procName, NULL);
    pixaa->n = 0;
    pixaa->nalloc = n;
    
    if ((pixaa->pixa = (PIXA **)CALLOC(n, sizeof(PIXA *))) == NULL)
	return (PIXAA *)ERROR_PTR("pixa ptrs not made", procName, NULL);

    return pixaa;
}


/*!
 *  pixaaDestroy()
 *
 *       Input:  &pixaa <to be nulled>
 *       Return:  void
 */
void
pixaaDestroy(PIXAA  **ppixaa)
{
INT32   i;
PIXAA   *pixaa;

    PROCNAME("pixaaDestroy");

    if (ppixaa == NULL) {
        WARNING("ptr address is NULL!", procName);
	return;
    }

    if ((pixaa = *ppixaa) == NULL)
	return;

    for (i = 0; i < pixaa->n; i++)
	pixaDestroy(&pixaa->pixa[i]);
    FREE((void *)pixaa->pixa);

    FREE((void *)pixaa);
    *ppixaa = NULL;

    return;
}


/*---------------------------------------------------------------------*
 *                             Pixaa addition                          *
 *---------------------------------------------------------------------*/
/*!
 *  pixaaAddPixa()
 *
 *       Input:  pixaa
 *               pixa  (to be added)
 *               copyflag (INSERT, COPY, CLONE)
 */
INT32
pixaaAddPixa(PIXAA   *pixaa,
             PIXA    *pixa,
             INT32    copyflag)
{
INT32    n;
PIXA    *pixac;

    PROCNAME("pixaaAddPixa");

    if (!pixaa)
	return ERROR_INT("pixaa not defined", procName, 1);
    if (!pixa)
	return ERROR_INT("pixa not defined", procName, 1);
    if (copyflag != INSERT && copyflag != COPY && copyflag != CLONE)
	return ERROR_INT("copyflag must be in {1,2,3}", procName, 1);

    if (copyflag == INSERT)
        pixac = pixa;
    else {
	if ((pixac = pixaCopy(pixa, copyflag)) == NULL)
	    return ERROR_INT("pixac not made", procName, 1);
    }

    n = pixaaGetCount(pixaa);
    if (n >= pixaa->nalloc)
	pixaaExtendArray(pixaa);
    pixaa->pixa[n] = pixac;
    pixaa->n++;

    return 0;
}


/*!
 *  pixaaExtendArray()
 *
 *       Input:  pixaa
 *       Return: 0 if OK; 1 on error
 */
INT32
pixaaExtendArray(PIXAA   *pixaa)
{

    PROCNAME("pixaaExtendArray");

    if (!pixaa)
	return ERROR_INT("pixaa not defined", procName, 1);

    if ((pixaa->pixa = (PIXA **)reallocNew((void **)&pixaa->pixa,
                                4 * pixaa->nalloc,
				8 * pixaa->nalloc)) == NULL)
	return ERROR_INT("new ptr array not returned", procName, 1);

    pixaa->nalloc = 2 * pixaa->nalloc;
    return 0;
}



/*---------------------------------------------------------------------*
 *                            Pixaa accessors                          *
 *---------------------------------------------------------------------*/
/*!
 *  pixaaGetCount()
 *
 *       Input:  pixaa
 *       Return: count, or 0 if no pixaa
 */
INT32
pixaaGetCount(PIXAA  *pixaa)
{

    PROCNAME("pixaaGetCount");

    if (!pixaa)
	return ERROR_INT("pixaa not defined", procName, 0);

    return pixaa->n;
}


/*!
 *  pixaaGetPixa()
 *
 *       Input:  pixaa
 *               index  (to the index-th pixa)
 *               accesstype  (COPY, CLONE, NEWREF)
 *       Return: pixa, or null on error
 *
 *       Usage note: COPY makes a new pixa with a copy of every pix
 *                   CLONE makes a new pixa with a clone of every pix
 *                   NEWREF just makes a new reference to the pixa,
 *                      and bumps the counter.  You would use this,
 *                      for example, when you need to extract some data
 *                      from a pix within a pixa within a pixaa.
 *                   In all cases, you must invoke destroyPixa() on the
 *                   returned pixa!
 */
PIXA *
pixaaGetPixa(PIXAA   *pixaa,
             INT32   index,
	     INT32   accesstype)
{
PIXA     *pixa;

    PROCNAME("pixaaGetPixa");

    if (!pixaa)
	return (PIXA *)ERROR_PTR("pixaa not defined", procName, NULL);
    if (index < 0 || index >= pixaa->n)
	return (PIXA *)ERROR_PTR("index not valid", procName, NULL);
    if (accesstype != COPY && accesstype != CLONE && accesstype != NEWREF)
	return (PIXA *)ERROR_PTR("accesstype must be in {0,1,2}",
	    procName, NULL);

    if (accesstype == NEWREF) {
        if ((pixa = pixaa->pixa[index]) == NULL)
	    return (PIXA *)ERROR_PTR("no pixa[index] !!", procName, NULL);
	pixaChangeRefcount(pixa, 1);
	return pixa;
    }
    else
	return pixaCopy(pixaa->pixa[index], accesstype);
}



/*---------------------------------------------------------------------*
 *                       Pixa and Pixaa Display                        *
 *---------------------------------------------------------------------*/
/*!
 *  pixaDisplay()
 *
 *         Input:  pixa
 *                 w, h (if set to 0, determines the size from the
 *                       b.b. of the components in pixa)
 */
PIX *
pixaDisplay(PIXA  *pixa,
            INT32  w,
	    INT32  h)
{
INT32    i, n, d;
BOX     *box;
BOXA    *boxa;
PIX     *pixt, *pixd;

    PROCNAME("pixaDisplay");

    if (!pixa)
	return (PIX *)ERROR_PTR("pixa not defined", procName, NULL);
    
    n = pixaGetCount(pixa);
    if (n == 0)
	return (PIX *)ERROR_PTR("no components", procName, NULL);

	/* if w and h not input, determine the minimum size required
	 * to contain the origin and all c.c. */
    if (w == 0 || h == 0)
	boxaGetExtent(pixa->boxa, &w, &h);

	/* Use the first pix in pixa to determine the depth.  */
    pixt = pixaGetPix(pixa, 0, CLONE);
    d = pixGetDepth(pixt);
    pixDestroy(&pixt);

    if ((pixd = pixCreate(w, h, d)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    for (i = 0; i < n; i++) {
	box = pixaGetBox(pixa, i, CLONE);
	pixt = pixaGetPix(pixa, i, CLONE);
	pixRasterop(pixd, box->x, box->y, box->w, box->h,
		    PIX_PAINT, pixt, 0, 0);
	boxDestroy(&box);
	pixDestroy(&pixt);
    }

    return pixd;
}


/*!
 *  pixaDisplayOnLattice()
 *
 *      Input:  pixa
 *              xspace
 *              yspace
 *
 *      Usage: this is used to save a pixa to a single image file,
 *             by putting the components on a rectangular lattice
 */
PIX *
pixaDisplayOnLattice(PIXA  *pixa,
                     INT32  xspace,
                     INT32  yspace)
{
INT32    n, nw, nh, w, h, d, wt, ht;
INT32    index, i, j;
PIX     *pixt, *pixd;

    PROCNAME("pixaDisplayOnLattice");

    if (!pixa)
	return (PIX *)ERROR_PTR("pixa not defined", procName, NULL);
    
    if ((n = pixaGetCount(pixa)) == 0)
	return (PIX *)ERROR_PTR("no components", procName, NULL);
    nw = (INT32)sqrt(n);
    nh = (n + nw - 1) / nw;
    w = xspace * nw;
    h = yspace * nh;

	/* Use the first pix in pixa to determine the depth.  */
    pixt = pixaGetPix(pixa, 0, CLONE);
    d = pixGetDepth(pixt);
    pixDestroy(&pixt);

    if ((pixd = pixCreate(w, h, d)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    
    index = 0;
    for (i = 0; i < nh; i++) {
	for (j = 0; j < nw && index < n; j++, index++) {
	    pixt = pixaGetPix(pixa, index, CLONE);
	    wt = pixGetWidth(pixt);
	    ht = pixGetHeight(pixt);
	    if (wt > xspace || ht > yspace) {
		fprintf(stderr, "pix(%d) omitted; size %dx%d\n", index, wt, ht);
		pixDestroy(&pixt);
		continue;
	    }
	    pixRasterop(pixd, j * xspace, i * yspace, wt, ht, 
			PIX_PAINT, pixt, 0, 0);
	    pixDestroy(&pixt);
	}
    }

    return pixd;
}


/*!
 *  pixaaDisplay()
 *
 *         Input:  pixaa
 *                 w, h (if set to 0, determines the size from the
 *                       b.b. of the components in pixaa)
 *         Return: pix, or null on error
 *
 *         Action: each pix of the pixaa is displayed at the
 *                 location given by its box
 */
PIX *
pixaaDisplay(PIXAA  *pixaa,
             INT32   w,
	     INT32   h)
{
INT32    i, j, n, na, d, wmax, hmax;
BOX     *box;
BOXA    *boxa;
PIX     *pixt, *pixd;
PIXA    *pixa;

    PROCNAME("pixaaDisplay");

    if (!pixaa)
	return (PIX *)ERROR_PTR("pixaa not defined", procName, NULL);
    
    n = pixaaGetCount(pixaa);
    if (n == 0)
	return (PIX *)ERROR_PTR("no components", procName, NULL);

	/* if w and h not input, determine the minimum size required
	 * to contain the origin and all c.c. */
    if (w == 0 || h == 0) {
	wmax = hmax = 0;
	for (i = 0; i < n; i++) {
	    pixa = pixaaGetPixa(pixaa, i, NEWREF);
	    boxaGetExtent(pixa->boxa, &w, &h);
	    wmax = MAX(wmax, w);
	    hmax = MAX(hmax, h);
	    pixaDestroy(&pixa);
	}
	w = wmax;
	h = hmax;
    }

	/* get depth from first pix */
    pixa = pixaaGetPixa(pixaa, 0, NEWREF);
    pixt = pixaGetPix(pixa, 0, CLONE);
    d = pixGetDepth(pixt);
    pixaDestroy(&pixa);
    pixDestroy(&pixt);

    if ((pixd = pixCreate(w, h, d)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    
    for (i = 0; i < n; i++) {
	pixa = pixaaGetPixa(pixaa, i, NEWREF);
	na = pixaGetCount(pixa);
	for (j = 0; j < na; j++) {
	    box = pixaGetBox(pixa, j, CLONE);
	    pixt = pixaGetPix(pixa, j, CLONE);
	    pixRasterop(pixd, box->x, box->y, box->w, box->h,
			PIX_PAINT, pixt, 0, 0);
	    boxDestroy(&box);
	    pixDestroy(&pixt);
	}
	pixaDestroy(&pixa);
    }

    return pixd;
}


/*!
 *  pixaaDisplayByPixa()
 *
 *         Input:  pixaa
 *                 xspace between pix in pixa
 *                 yspace between pixa
 *                 max width of output pix
 *         Return: pix, or null on error
 *
 *         Action: displays each pixa on a line (or set of lines),
 *                 in order from top to bottom.  Within each pixa,
 *                 the pix are displayed in order from left to right.
 *
 *         Note: The size of each pix in each pixa is assumed to be
 *               approximately equal to the size of the first pix in
 *               the pixa.  If this assumption is not correct, this
 *               function will not work properly.
 */
PIX *
pixaaDisplayByPixa(PIXAA  *pixaa,
                   INT32   xspace,
	           INT32   yspace,
		   INT32   maxw)
{
INT32    i, j, npixa, npix;
INT32    width, height, depth, nlines, lwidth;
INT32    x, y, w, h, w0, h0;
PIX     *pixt, *pixd;
PIXA    *pixa;

    PROCNAME("pixaaDisplayByPixa");

    if (!pixaa)
	return (PIX *)ERROR_PTR("pixaa not defined", procName, NULL);
    
    if ((npixa = pixaaGetCount(pixaa)) == 0)
	return (PIX *)ERROR_PTR("no components", procName, NULL);

	/* get size of output pix.  The width is the minimum of the
	 * maxw and the largest pixa line width.  The height is whatever
	 * it needs to be to accommodate all pixa. */
    height = 2 * yspace;
    width = 0;
    for (i = 0; i < npixa; i++) {
        pixa = pixaaGetPixa(pixaa, i, NEWREF);
	npix = pixaGetCount(pixa);
	pixt = pixaGetPix(pixa, 0, CLONE);
	if (i == 0)
	    depth = pixGetDepth(pixt);
	w = pixGetWidth(pixt);
	lwidth = npix * (w + xspace);
	nlines = (lwidth + maxw - 1) / maxw;
	if (nlines > 1)
	    width = maxw;
	else
	    width = MAX(lwidth, width);
	height += nlines * (pixGetHeight(pixt) + yspace);
	pixDestroy(&pixt);
	pixaDestroy(&pixa);
    }

    if ((pixd = pixCreate(width, height, depth)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

        /* now layout the pix by pixa */
    y = yspace;
    for (i = 0; i < npixa; i++) {
	x = 0;
	pixa = pixaaGetPixa(pixaa, i, NEWREF);
	npix = pixaGetCount(pixa);
	for (j = 0; j < npix; j++) {
	    pixt = pixaGetPix(pixa, j, CLONE);
	    if (j == 0) {
	        w0 = pixGetWidth(pixt);
	        h0 = pixGetHeight(pixt);
	    }
	    w = pixGetWidth(pixt);
	    if (width == maxw && x + w >= maxw) {
	        x = 0;
		y += h0 + yspace;
	    }
	    h = pixGetHeight(pixt);
	    pixRasterop(pixd, x, y, w, h, PIX_PAINT, pixt, 0, 0);
	    pixDestroy(&pixt);
	    x += w0 + xspace;
	}
	y += h0 + yspace;
	pixaDestroy(&pixa);
    }

    return pixd;
}



/*---------------------------------------------------------------------*
 *                          Pixa serialized I/O                        *
 *---------------------------------------------------------------------*/
/*!
 *  pixaWriteStream()
 *
 *      Input: stream
 *             pixa
 *      Return: 0 if OK, 1 on error
 */
INT32
pixaWriteStream(FILE  *fp,
                PIXA  *pixa)
{
INT32     n, i;
PIX      *pix;
    PROCNAME("boxWriteStream");

    if (!fp)
	return ERROR_INT("stream not defined", procName, 1);
    if (!pixa)
	return ERROR_INT("pixa not defined", procName, 1);

    n = pixaGetCount(pixa);
    fprintf(fp, "\npixa: number of pix = %d\n", n);
    boxaWriteStream(fp, pixa->boxa);
    for (i = 0; i < n; i++) {
	if ((pix = pixaGetPix(pixa, i, CLONE)) == NULL)
	    return ERROR_INT("pix not found", procName, 1);
	fprintf(fp, " pix[%d]: xres = %d, yres = %d\n",
		i, pix->xres, pix->yres);
	pixWriteStreamPng(fp, pix);
	pixDestroy(&pix);
    }
    return 0;
}


/*!
 *  pixaReadStream()
 *
 *      Input: stream
 *      Return: pixa, or null on error
 */
PIXA *
pixaReadStream(FILE  *fp)
{
INT32    n, i, xres, yres;
INT32    ignore;
BOXA    *boxa;
PIX     *pix;
PIXA    *pixa;

    PROCNAME("pixaReadStream");

    if (!fp)
	return (PIXA *)ERROR_PTR("stream not defined", procName, NULL);

    if (fscanf(fp, "\npixa: number of pix = %d\n", &n) != 1)
	return (PIXA *)ERROR_PTR("not a pixa file", procName, NULL);
    if ((pixa = pixaCreate(n)) == NULL)
	return (PIXA *)ERROR_PTR("pixa not made", procName, NULL);
    if ((boxa = boxaReadStream(fp)) == NULL)
	return (PIXA *)ERROR_PTR("boxa not made", procName, NULL);
    boxaDestroy(&pixa->boxa);
    pixa->boxa = boxa;

    for (i = 0; i < n; i++) {
	if ((fscanf(fp, " pix[%d]: xres = %d, yres = %d\n",
	      &ignore, &xres, &yres)) != 3)
	    return (PIXA *)ERROR_PTR("res reading", procName, NULL);
	if ((pix = pixReadStreamPng(fp)) == NULL)
	    return (PIXA *)ERROR_PTR("pix not read", procName, NULL);
	pixSetXRes(pix, xres);
	pixSetYRes(pix, yres);
	pixaAddPix(pixa, pix, INSERT);
    }

    return pixa;
}

