/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/

/*
 *  pngio.c
 *                     
 *          PIX        *pixReadStreamPng()
 *          INT32       pixWriteStreamPng()
 *
 *
 *    Documentation: libpng.txt and example.c
 *
 *    On input (decompression from file), palette color images
 *    are read into an 8 bpp Pix with a colormap, and 24 bpp
 *    3 component color images are read into a 32 bpp Pix with
 *    rgb samples.  On output (compression to file), palette color
 *    images are written as 8 bpp with the colormap, and 32 bpp
 *    full color images are written compressed as a set of three
 *    8 bpp (rgb) images.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "allheaders.h"

#ifndef  NO_CONSOLE_IO
#define  DEBUG     0
#endif  /* ~NO_CONSOLE_IO */


/*!
 *  pixReadStreamPng()
 *
 *      Input:  stream
 *      Return: pix, or null on error
 *
 *      Usage note:
 *         If called from pixReadStream(), the stream
 *         is positioned at the beginning of the file.
 *         If you want to do sequential reads of png format
 *         images from a stream, use pixReadStreamPng()
 *                 
 */
PIX *
pixReadStreamPng(FILE  *fp)
{
INT32		      i, j, k;
INT32                 wpl, d, spp, cindex;
UINT32                png_transforms;
UINT32               *data, *line, *ppixel;
int                   num_palette;
png_byte              bit_depth, color_type, channels;
png_uint_32           w, h, rowbytes;
png_bytep             rowptr;
png_bytep            *row_pointers;
png_structp           png_ptr;
png_infop             info_ptr, end_info;
png_colorp            palette;
GIFColorTableEntry   *gifcmap;
PIX 		     *pix;
PIXCMAP              *cmap;


    PROCNAME("pixReadStreamPng");

    if (!fp)
        return (PIX *)ERROR_PTR("fp not defined", procName, NULL);
    pix = NULL;

/*    rewind(fp); */

        /* allocate the 3 data structures */
    if ((png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
                   (png_voidp)NULL, NULL, NULL)) == NULL)
        return (PIX *)ERROR_PTR("png_ptr not made", procName, NULL);

    if ((info_ptr = png_create_info_struct(png_ptr)) == NULL) {
	png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
        return (PIX *)ERROR_PTR("info_ptr not made", procName, NULL);
    }

    if ((end_info = png_create_info_struct(png_ptr)) == NULL) {
	png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
        return (PIX *)ERROR_PTR("end_info not made", procName, NULL);
    }

	/* set up png setjmp error handling */
    if (setjmp(png_jmpbuf(png_ptr))) {
	png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
        return (PIX *)ERROR_PTR("internal png error", procName, NULL);
    }

    png_init_io(png_ptr, fp);

	/* remove alpha channel; strip 16 --> 8 bit depth; but
	 * DO NOT invert binary using PNG_TRANSFORM_INVERT_MONO! */
    png_transforms = PNG_TRANSFORM_STRIP_16 | PNG_TRANSFORM_STRIP_ALPHA;
    png_read_png(png_ptr, info_ptr, png_transforms, NULL);

    row_pointers = png_get_rows(png_ptr, info_ptr);
    w = png_get_image_width(png_ptr, info_ptr);
    h = png_get_image_height(png_ptr, info_ptr);
    bit_depth = png_get_bit_depth(png_ptr, info_ptr);
    rowbytes = png_get_rowbytes(png_ptr, info_ptr);
    color_type = png_get_color_type(png_ptr, info_ptr);
    channels = png_get_channels(png_ptr, info_ptr);

    spp = channels;

    if (spp == 1)
	d = bit_depth;
    else if (spp == 2) {
	d = 2 * bit_depth;
	WARNING("there shouldn't be 2 spp!", procName);
    }
    else if (spp == 3)
	d = 4 * bit_depth;
    else  {  /* spp == 4 */
	d = 4 * bit_depth;
	WARNING("there shouldn't be 4 spp!", procName);
    }

	/* remove when this is implemented for all bit_depths */
    if (spp == 3 && bit_depth != 8) {
	fprintf(stderr, "Help: spp = 3 and bit_depth = %d\n!!", bit_depth);
	return (PIX *)ERROR_PTR("not implemented for this depth",
	    procName, NULL);
    }

    if (color_type == PNG_COLOR_TYPE_PALETTE ||
        color_type == PNG_COLOR_MASK_PALETTE)
    {    /* generate a colormap */
	png_get_PLTE(png_ptr, info_ptr, &palette, &num_palette);

        if ((gifcmap =
	    (GIFColorTableEntry *)CALLOC(sizeof(GIFColorTableEntry),
	                                 num_palette)) == NULL)
        {
	    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
	    return (PIX *)ERROR_PTR( "gifcmap not made", procName, NULL);
	}
	for (cindex = 0; cindex < num_palette; cindex++) {
	    gifcmap[cindex].cmRed = palette[cindex].red;
	    gifcmap[cindex].cmGreen = palette[cindex].green;
	    gifcmap[cindex].cmBlue = palette[cindex].blue;
	}

	if ((cmap = (PIXCMAP *)CALLOC(1, sizeof(PIXCMAP))) == NULL) {
	    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
	    return (PIX *)ERROR_PTR("cmap not made", procName, NULL);
	}
	cmap->type = CM_GIF;
	cmap->array  = gifcmap;
	cmap->size  = num_palette * sizeof(GIFColorTableEntry);
    }
    else
	cmap = NULL;

    if ((pix = pixCreate(w, h, d)) == NULL) {
	png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
	return (PIX *)ERROR_PTR("pix not made", procName, NULL);
    }
    wpl = pixGetWpl(pix);
    data = pixGetData(pix);
    pixSetColormap(pix, cmap);

    if (spp == 1) {   /* copy straight from buffer to pix */
	for (i = 0; i < h; i++) {
	    line = data + i * wpl;
	    rowptr = row_pointers[i];
	    for (j = 0; j < rowbytes; j++) {
		SET_DATA_BYTE(line, j, rowptr[j]);
	    }
	}
    }
    else  {   /* spp == 3 */
	for (i = 0; i < h; i++) {
	    ppixel = data + i * wpl;
	    rowptr = row_pointers[i];
            for (j = k = 0; j < w; j++)
	    {
		SET_DATA_BYTE(ppixel, 0, rowptr[k++]);      /* R */
		SET_DATA_BYTE(ppixel, 1, rowptr[k++]);      /* G */
		SET_DATA_BYTE(ppixel, 2, rowptr[k++]);      /* B */
		ppixel++;
	    }
	}
    }

#if  DEBUG
    if (cmap) {
	for (i = 0; i < 16; i++) {
	    fprintf(stderr, "[%d] = %d\n", i,
		   ((UINT8 *)(cmap->array))[i]);
	}
    }
#endif  /* DEBUG */

	/* 
	 * If there is no colormap, PNG defines black = 0 and
	 * white = 1 by default for binary monochrome.  Therefore,
	 * since we use the opposite definition, we must invert
	 * the image in either of these cases:
	 *    (i) there is no colormap (default)
	 *    (ii) there is a colormap which defines black to
	 *         be 0 and white to be 1.
	 * We cannot use the PNG_TRANSFORM_INVERT_MONO flag
	 * because that flag (since 1.0.9) inverts 8 bpp
	 * grayscale as well, which we don't want to do.
	 * (It also doesn't work if there is a colormap.)
	 * If there is a colormap that defines black = 1 and
	 * white = 0, we don't need to do anything.
	 * 
	 * How do we check the polarity of the colormap?
	 * The gif colormap determines the values of black and
	 * white pixels in the following way:
	 *     if black = 1 (255), white = 0
	 *          255, 255, 255, 0, 0, 0
	 *     if black = 0, white = 1 (255)
	 *          0, 0, 0, 255, 255, 255
	 * So we test the first byte to see if it is 0; if so, we
	 * must invert the data.
	 */
    if (d == 1 && (!cmap || (cmap && ((UINT8 *)(cmap->array))[0] == 0x0))) {
#if DEBUG
	fprintf(stderr, "Inverting binary data on png read\n");
#endif  /* DEBUG */
	pixInvert(pix, pix);
    }

    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);

    return pix;
}


/*!
 *  pixWriteStreamPng()
 *
 *      Input:  stream, pix
 *      Return: 0 if OK; 1 on error
 *
 *      Usage note:
 *         If called from pixWriteStream(), the stream
 *         is positioned at the beginning of the file.
 *         If you want to do sequential writes of png format
 *         images to a stream, use pixWriteStreamPng()
 *                 
 */
INT32
pixWriteStreamPng(FILE  *fp,
                  PIX   *pix)
{
UINT8                *rmap, *gmap, *bmap;
INT32		      i, j, k;
INT32                 wpl, d, cmflag;
UINT32               *data, *ppixel;
INT32                 numcolors;
png_byte              bit_depth, color_type;
png_uint_32           w, h;
png_bytep            *row_pointers;
png_bytep             rowbuffer;
png_structp           png_ptr;
png_infop             info_ptr;
png_colorp            palette;
PIXCMAP              *cmap;

    PROCNAME("pixWriteStreamPng");

    if (!fp)
        return ERROR_INT("stream not open", procName, 1);
    if (!pix)
        return ERROR_INT("pix not defined", procName, 1);

/*    rewind(fp); */

        /* allocate the 2 data structures */
    if ((png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
                   (png_voidp)NULL, NULL, NULL)) == NULL)
        return ERROR_INT("png_ptr not made", procName, 1);

    if ((info_ptr = png_create_info_struct(png_ptr)) == NULL) {
	png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
        return ERROR_INT("info_ptr not made", procName, 1);
    }

	/* set up png setjmp error handling */
    if (setjmp(png_jmpbuf(png_ptr))) {
	png_destroy_write_struct(&png_ptr, &info_ptr);
        return ERROR_INT("internal png error", procName, 1);
    }

    png_init_io(png_ptr, fp);

    w = pixGetWidth(pix);
    h = pixGetHeight(pix);
    d = pixGetDepth(pix);
    wpl = pixGetWpl(pix);
    data = pixGetData(pix);
    if (cmap = pixGetColormap(pix))
	cmflag = 1;
    else
	cmflag = 0;
    if (d == 32) {
	bit_depth = 8;
	color_type = PNG_COLOR_TYPE_RGB;
	cmflag = 0;  /* ignore if it exists */
    }
    else {
	bit_depth = d;
	color_type = PNG_COLOR_TYPE_GRAY;
    }
    if (cmflag)
	color_type = PNG_COLOR_TYPE_PALETTE;

#if  DEBUG
    fprintf(stderr, "cmflag = %d, bit_depth = %d, color_type = %d\n",
            cmflag, bit_depth, color_type);
#endif  /* DEBUG */

    png_set_IHDR(png_ptr, info_ptr, w, h, bit_depth, color_type,
                 PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE,
		 PNG_FILTER_TYPE_BASE);

	/* make useful colormap arrays */
    if (cmflag) {
	if (colormapToArrays(cmap, &rmap, &gmap, &bmap, &numcolors))
	    return ERROR_INT("unidentified colormap type", procName, 1);
    }

	/* make and save the palette */
    if (cmflag) {
	if ((palette = (png_colorp)(CALLOC(numcolors, sizeof(png_color))))
		== NULL)
	    return ERROR_INT("palette not made", procName, 1);

	for (i = 0; i < numcolors; i++) {
	    palette[i].red = rmap[i];
	    palette[i].green = gmap[i];
	    palette[i].blue = bmap[i];
	}

	png_set_PLTE(png_ptr, info_ptr, palette, (int)numcolors);

	FREE((void *)rmap);
	FREE((void *)gmap);
	FREE((void *)bmap);
    }

	/* write header and palette info */
    png_write_info(png_ptr, info_ptr);
	    
    if (d != 32)   /* not 24 bit color */
    {
	    /* make and store array of image row pointers */
	if ((row_pointers = (png_bytep *)CALLOC(h, sizeof(png_bytep))) == NULL)
	    return ERROR_INT("row-pointers not made", procName, 1);
	for (i = 0; i < h; i++)
	    row_pointers[i] = (png_bytep)(data + i * wpl);
	png_set_rows(png_ptr, info_ptr, row_pointers);

	    /* for a binary image, there are two conditions in
	     * which you must invert the data for write png:
	     *    (a) no colormap
	     *    (b) colormap with BLACK set to 0
	     * png writes binary with BLACK = 0, unless contradicted
	     * by a colormap.  If the colormap has BLACK = "1"
	     * (typ. about 255), do not invert the data.  If there
	     * is no colormap, you must invert the data to store
	     * in default BLACK = 0 state.
	     */
	if (d == 1 && (!cmap || (cmap && ((UINT8 *)(cmap->array))[0] == 0x0))) {
#if DEBUG
	    fprintf(stderr, "Inverting binary data on png write\n");
#endif  /* DEBUG */
	    pixInvert(pix, pix);
	}

	pixEndianByteSwap(pix);

	png_write_image(png_ptr, row_pointers);

	pixEndianByteSwap(pix);

	if (d == 1 && (!cmap || (cmap && ((UINT8 *)(cmap->array))[0] == 0x0)))
	    pixInvert(pix, pix);

	png_write_end(png_ptr, info_ptr);

	if (cmflag)
	    FREE((void *)palette);
	FREE((void *)row_pointers);
	png_destroy_write_struct(&png_ptr, &info_ptr);
	return 0;
    }

	/* 24 bit color; write a row at a time */
    if ((rowbuffer = (png_bytep)CALLOC(w, 3)) == NULL)
	return ERROR_INT("rowbuffer not made", procName, 1);
    for (i = 0; i < h; i++)
    {
	ppixel = data + i * wpl;
	for (j = k = 0; j < w; j++) {
	    rowbuffer[k++] = GET_DATA_BYTE(ppixel, 0);
	    rowbuffer[k++] = GET_DATA_BYTE(ppixel, 1);
	    rowbuffer[k++] = GET_DATA_BYTE(ppixel, 2);
	    ppixel++;
	}

	png_write_rows(png_ptr, &rowbuffer, 1);
    }

    png_write_end(png_ptr, info_ptr);

    if (cmflag)
	FREE((void *)palette);
    FREE((void *)rowbuffer);
    png_destroy_write_struct(&png_ptr, &info_ptr);
    return 0;

}

