/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/

/*
 *  psio.c
 *                     
 *          INT32       pixWriteStreamPS()
 *          char       *pixWriteStringPS()
 *
 *          INT32       convertByteToHexAscii()
 *
 *     This is a PostScript "device driver" for wrapping images
 *     in PostScript.  The images can be rendered by a PostScript
 *     interpreter, such as gs or the embedded interpreter in a
 *     PostScript printer.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "allheaders.h"

#define  BUF_SIZE              512
#define  DEFAULT_PRINTER_RES   300   /* default ppi for printing */
#define  MIN_RES               25
#define  MAX_RES               3000


/*!
 *  pixWriteStreamPS()
 *
 *      Input:  stream
 *              pix
 *              box  (<optional>)
 *              res  (can use 0 for default of 300 ppi)
 *              scale (to prevent scaling, use either 1.0 or 0.0)
 *      Return: 0 if OK; 1 on error
 *
 *      Action: writes image in PS format, optionally scaled,
 *              adjusted for the printer resolution, and with
 *              a bounding box.  For details on use of parameters,
 *              see pixWriteStringPS().
 */
INT32
pixWriteStreamPS(FILE    *fp,
                 PIX     *pix,
		 BOX     *box,
		 INT32    res,
		 FLOAT32  scale)
{
char    *pstring;
INT32    length;

    PROCNAME("pixWriteStreamPS");

    if (!fp)
        return (INT32)ERROR_INT("stream not open", procName, 1);
    if (!pix)
        return (INT32)ERROR_INT("pix not defined", procName, 1);

    pstring = pixWriteStringPS(pix, box, res, scale);
    length = strlen(pstring);
    fwrite(pstring, 1, length, fp);
    return 0;
}


/*!
 *  pixWriteStringPS()
 *
 *      Input:  pix:  1 bpp, 8 bpp (colormap ignored), 32 bpp (RGB)
 *              box:  (a) If box == null, image is placed, optionally scaled,
 *                        in a standard b.b. at the center of the page.
 *                        This is to be used when another program like
 *                        TeX (through epsf) places the image.
 *                    (b) If box != null, image is placed without a
 *                        b.b. at the specified page location and with
 *                        (optional) scaling.  This is to be used when
 *                        you want to specify exactly where (and optionally
 *                        how big) you want the image to be.
 *                        Note that all coordinates are in PS convention,
 *                        with (0,0) at LL corner of the page:
 *                            (x,y)    location of LL corner of image, in mils.
 *                            (w,h)    scaled size, in mils.  Use 0 to
 *                                     scale with "scale" and "res" input.
 *              res:  resolution, in printer ppi.  Use 0 for default (300 ppi).
 *              scale: scale factor.  If no scaling is desired, use
 *                     either 1.0 or 0.0.   Scaling just resets the resolution
 *                     parameter; the actual scaling is done in the
 *                     interpreter at rendering time.  This is important:
 *                     it allows you to scale the image up without
 *                     increasing the file size.
 *
 *      Return: ps string if OK, or null on error
 *
 *      Usage notes:
 *          OK, this seems a bit complicated, because there are various
 *          ways to scale and not to scale.
 *          If you don't want any scaling at all:
 *              if you are using a box:
 *                  set w = 0, h = 0, and use scale = 1.0; it will print
 *                  each pixel unscaled at printer resolution
 *              if you are not using a box:
 *                  set scale = 1.0; it will print at printer resolution
 *          If you want the image to be a certain size in inches:
 *              you must use a box and set the box (w,h) in mils
 *          If you want the image to be scaled by a scale factor != 1.0:
 *              if you are using a box:
 *                  set w = 0, h = 0, and use the desired scale factor;
 *                  the higher the printer resolution, the smaller the
 *                  image will actually appear.
 *              if you are not using a box:
 *                  set the desired scale factor; the higher the printer
 *                  resolution, the smaller the image will actually appear.
 * 
 *                     
 *      Distance units:  The interface distances are in milli-inches.
 *                       In addition, we use 3 different units internally:
 *                           pixels  (units of 1/res inch)
 *                           printer pts (units of 1/72 inch)
 *                           inches
 */
char *
pixWriteStringPS(PIX      *pix,
                 BOX      *box,
		 INT32     res,
		 FLOAT32   scale)
{
char       nib1, nib2;
char       bigbuf[BUF_SIZE];
char      *hexdata, *pstring;
UINT8      byteval;
INT32      i, j, k, d, wpix, hpix, wpt, hpt, xpt, ypt;
INT32      wpl, psbpl, hexbytes, boxflag, sampledepth;
UINT32    *line, *data;
FLOAT32    winch, hinch, xinch, yinch;
SARRAY    *sa;


    PROCNAME("pixWriteStringPS");

    if (!pix)
        return (char *)ERROR_PTR("pix not defined", procName, NULL);

    d = pixGetDepth(pix);
    if (d != 1 && d!= 8 && d != 32)
        return (char *)ERROR_PTR("d must be in {1, 8, 32}", procName, NULL);

    if (!box)
	boxflag = 0;  /* no scaling; b.b. at center */
    else
	boxflag = 1;  /* no b.b., specify placement and optional scaling */

    if (res == 0)
	res = DEFAULT_PRINTER_RES;

	/* scale using the PS interpreter in the printer */
    if (scale == 0.0)
	scale = 1.0;
    if (scale != 1.0)
	res = (INT32)((FLOAT32)res / scale);

    if (res < MIN_RES || res > MAX_RES) {
	WARNING_INT("res %d out of bounds; using default", procName, res);
	res = DEFAULT_PRINTER_RES;
    }

    wpix = pixGetWidth(pix);
    hpix = pixGetHeight(pix);
    if (boxflag == 0) {
	winch = (FLOAT32)wpix / (FLOAT32)res;
	hinch = (FLOAT32)hpix / (FLOAT32)res;
	xinch = (8.5 - winch) / 2.;
	yinch = (11.0 - hinch) / 2.;
    }
    else {   /* use box */
	if (box->w == 0)
	    winch = (FLOAT32)wpix / (FLOAT32)res;
	else
	    winch = (FLOAT32)box->w / 1000.;
	if (box->h == 0)
	    hinch = (FLOAT32)hpix / (FLOAT32)res;
	else
	    hinch = (FLOAT32)box->h / 1000.;
	xinch = (FLOAT32)box->x / 1000.;
	yinch = (FLOAT32)box->y / 1000.;
    }

    if (xinch < 0)
	WARNING("left edge < 0.0 inch", procName);
    if (xinch + winch > 8.5)
	WARNING("right edge > 8.5 inch", procName);
    if (yinch < 0.0)
	WARNING("bottom edge < 0.0 inch", procName);
    if (yinch + hinch > 11.0)
	WARNING("top edge > 11.0 inch", procName);

    wpt = (INT32)(72. * winch);
    hpt = (INT32)(72. * hinch);
    xpt = (INT32)(72. * xinch);
    ypt = (INT32)(72. * yinch);
    
    if (d == 1)
	sampledepth = 1;
    else  /* d == 8 || d == 32 */
	sampledepth = 8;

	/* convert image data to hex string */
    wpl = pixGetWpl(pix);
    if (d == 1 || d == 8)
	psbpl = (wpix * d + 7) / 8;   /* packed to byte boundary */
    else /* d == 32 */
	psbpl = 3 * wpix;   /* packed to byte boundary */
    data = pixGetData(pix);
    hexbytes = 2 * psbpl * hpix;  /* size of ps hex array */
    if ((hexdata = (char *)CALLOC(hexbytes + 1, sizeof(char))) == NULL)
        return (char *)ERROR_PTR("hexdata not made", procName, NULL);
    if (d == 1 || d == 8) {
	for (i = 0, k = 0; i < hpix; i++) {
	    line = data + i * wpl;
	    for (j = 0; j < psbpl; j++) {
		byteval = GET_DATA_BYTE(line, j);
		convertByteToHexAscii(byteval, &nib1, &nib2);
		hexdata[k++] = nib1;
		hexdata[k++] = nib2;
	    }
	}
    }
    else  {  /* d == 32; hexdata bytes packed RGBRGB..., 2 per sample */
	for (i = 0, k = 0; i < hpix; i++) {
	    line = data + i * wpl;
	    for (j = 0; j < wpix; j++) {
		byteval = GET_DATA_BYTE(line + j, 0);  /* red */
		convertByteToHexAscii(byteval, &nib1, &nib2);
		hexdata[k++] = nib1;
		hexdata[k++] = nib2;
		byteval = GET_DATA_BYTE(line + j, 1);  /* green */
		convertByteToHexAscii(byteval, &nib1, &nib2);
		hexdata[k++] = nib1;
		hexdata[k++] = nib2;
		byteval = GET_DATA_BYTE(line + j, 2);  /* blue */
		convertByteToHexAscii(byteval, &nib1, &nib2);
		hexdata[k++] = nib1;
		hexdata[k++] = nib2;
	    }
	}
    }
    hexdata[k] = '\0';

    if ((sa = sarrayCreate(0)) == NULL)
        return (char *)ERROR_PTR("sa not made", procName, NULL);

    sarrayAddString(sa, "%!Adobe-PS", 1);
    if (boxflag == 0) {
	sprintf(bigbuf,
	    "%%%%BoundingBox: %d %d %d %d", xpt, ypt, xpt + wpt, ypt + hpt);
	sarrayAddString(sa, bigbuf, 1);
    }
    else    /* boxflag == 1 */
	sarrayAddString(sa, "gsave", 1);

    if (d == 1)
	sarrayAddString(sa, "{1 exch sub} settransfer    %invert binary", 1);

    sprintf(bigbuf, "/bpl %d string def         %%bpl as a string", psbpl);
    sarrayAddString(sa, bigbuf, 1);
    sprintf(bigbuf,
	"%d %d translate         %%set image origin in pts", xpt, ypt);
    sarrayAddString(sa, bigbuf, 1);
    sprintf(bigbuf,
	"%d %d scale             %%set image size in pts", wpt, hpt);
    sarrayAddString(sa, bigbuf, 1);
    sprintf(bigbuf,
	"%d %d %d                 %%image dimensions in pixels",
	    wpix, hpix, sampledepth);
    sarrayAddString(sa, bigbuf, 1);
    sprintf(bigbuf,
	"[%d %d %d %d %d %d]     %%mapping matrix: [wpix 0 0 -hpix 0 hpix]",
	       wpix, 0, 0, -hpix, 0, hpix);
    sarrayAddString(sa, bigbuf, 1);

    if (boxflag == 0) {
        if (d == 1 || d == 8)
	    sarrayAddString(sa, "{currentfile bpl readhexstring pop} image", 1);
	else  /* d == 32 */
	    sarrayAddString(sa,
	        "{currentfile bpl readhexstring pop} false 3 colorimage", 1);
    }
    else {  /* boxflag == 1 */
        if (d == 1 || d == 8)
	    sarrayAddString(sa,
	        "{currentfile bpl readhexstring pop} bind image",1);
	else  /* d == 32 */
	    sarrayAddString(sa,
	        "{currentfile bpl readhexstring pop} bind false 3 colorimage",1);
    }

    sarrayAddString(sa, hexdata, 0);

    if (boxflag == 0)
	sarrayAddString(sa, "\nshowpage", 1);
    else  /* boxflag == 1 */
	sarrayAddString(sa, "\ngrestore", 1);

    if ((pstring = sarrayToString(sa, 1)) == NULL)
        return (char *)ERROR_PTR("pstring not made", procName, NULL);

    sarrayDestroy(&sa);
    return pstring;
}


/*!
 *  convertByteToHexAscii()
 *
 *     Input:  byteval  (input byte)
 *             &nib1, &nib2  (<return> two hex ascii characters)
 *     Return: void
 */
void
convertByteToHexAscii(UINT8  byteval,
                      char  *pnib1,
		      char  *pnib2)
{
UINT8      nib;

    nib = byteval >> 4;
    if (nib < 10)
	*pnib1 = '0' + nib;
    else
	*pnib1 = 'a' + (nib - 10);
    nib = byteval & 0xf;
    if (nib < 10)
	*pnib2 = '0' + nib;
    else
	*pnib2 = 'a' + (nib - 10);

    return;
}

