/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  rotateam.c
 *
 *     Grayscale rotation (area mapped == interpolation)
 *
 *         Rotation about the image center:
 *                  PIX        *pixRotateAM()
 *                  PIX        *pixRotateAMColor()
 *                  PIX        *pixRotateAMGray()
 *
 *         Rotation (gray) about the UL corner of the image:
 *                  PIX        *pixRotateAMCorner()
 *
 *     Rotations are measured in radians; clockwise is positive.
 *
 *     The basic area mapping grayscale rotation works on 8 bpp images.
 *     The color rotation applies the grayscale rotation to each
 *     of the three color planes (rgb) in a full color 24 bpp
 *     image.  The full color images are stored in a 32 bpp Pix.
 *
 *     Area mapping works as follows.  For each dest
 *     pixel you find the 4 source pixels that it partially
 *     covers.  You then compute the dest pixel value as
 *     the area-weighted average of those 4 source pixels.
 *     We make two simplifying approximations:
 *
 *       -  For simplicity, compute the areas as if the dest
 *          pixel were translated but not rotated.
 *
 *       -  Compute area overlaps on a discrete sub-pixel grid.
 *          Because we are using 8 bpp images with 256 levels,
 *          it is convenient to break each pixel into a
 *          16x16 sub-pixel grid, and count the number of
 *          overlapped sub-pixels.
 *
 *     It is interesting to note that the digital filter that
 *     implements the area mapping algorithm for rotation
 *     is identical to the digital filter used for linear
 *     interpolation when arbitrarily scaling grayscale images.
 *
 *     The advantage of area mapping over pixel sampling
 *     in grayscale rotation is that it the former naturally
 *     blurs sharp edges ("anti-aliasing"), so that stair-step
 *     artifacts are not introduced.  The disadvantage is that
 *     it is significantly slower.
 * 
 */

#include <stdio.h>
#include <string.h>

#include "allheaders.h"

#define  VERY_SMALL_ANGLE       0.001    /* radians; approx. 0.05 degrees */



/*------------------------------------------------------------------*
 *                     Rotation about the center                    *
 *------------------------------------------------------------------*/
/*!
 *  pixRotateAM()
 *
 *         Input: pixs
 *                angle (radians; clockwise is positive)
 *         Return:  pixd, or null on error
 *
 */
PIX *
pixRotateAM(PIX     *pixs,
	    FLOAT32  angle)
{
INT32      d;
PIX    	  *pixt, *pixd;

    PROCNAME("pixRotateAM");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    d = pixGetDepth(pixs);

    if (d == 32)
	return pixRotateAMColor(pixs, angle);
    else if (pixGetColormap(pixs)) {  /* 2, 4 or 8 bpp */
	pixt = pixRemoveColormap(pixs, CONVERT_TO_FULL_COLOR);
	pixd = pixRotateAMColor(pixt, angle);
	pixDestroy(&pixt);
	return pixd;
    }
    else if (d == 8)
	return pixRotateAMGray(pixs, angle);
    else
        return (PIX *)ERROR_PTR("pixs must be 8 or 32 bpp", procName, NULL);
}


/*!
 *  pixRotateAMColor()
 *
 *         Input: pixs
 *                angle (radians; clockwise is positive)
 *         Return:  pixd, or null on error
 *
 *         Note: this rotates a color image about the image center.
 *               A positive angle gives a clockwise rotation.
 *
 */
PIX *
pixRotateAMColor(PIX     *pixs,
	         FLOAT32  angle)
{
PIX          *pixRed, *pixGreen, *pixBlue;
PIX          *pixRedRot, *pixGreenRot, *pixBlueRot;
PIX    	     *pixd;

    PROCNAME("pixRotateAMColor");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 32)
        return (PIX *)ERROR_PTR("pixs must be 32 bpp", procName, NULL);

    if (ABS(angle) < VERY_SMALL_ANGLE)
	return pixClone(pixs);

    pixRed = pixGetRGBComponent(pixs, COLOR_RED);
    pixRedRot = pixRotateAMGray(pixRed, angle);
    pixDestroy(&pixRed);
    pixGreen = pixGetRGBComponent(pixs, COLOR_GREEN);
    pixGreenRot = pixRotateAMGray(pixGreen, angle);
    pixDestroy(&pixGreen);
    pixBlue = pixGetRGBComponent(pixs, COLOR_BLUE);
    pixBlueRot = pixRotateAMGray(pixBlue, angle);
    pixDestroy(&pixBlue);

    if ((pixd = pixCreateRGBImage(pixRedRot, pixGreenRot, pixBlueRot)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    pixDestroy(&pixRedRot);
    pixDestroy(&pixGreenRot);
    pixDestroy(&pixBlueRot);

    return pixd;
}


/*!
 *  pixRotateAMGray()
 *
 *         Input: pixs
 *                angle (radians; clockwise is positive)
 *         Return:  pixd, or null on error
 *
 *         Note: this rotates the image about the image center.
 *               A positive angle gives a clockwise rotation.
 *
 */
PIX *
pixRotateAMGray(PIX     *pixs,
	        FLOAT32  angle)
{
INT32         w, h, wpls, wpld;
UINT32       *datas, *datad;
PIX    	     *pixd;

    PROCNAME("pixRotateAMGray");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 8)
        return (PIX *)ERROR_PTR("pixs must be 8 bpp", procName, NULL);

    if (ABS(angle) < VERY_SMALL_ANGLE)
	return pixClone(pixs);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);
    pixd = pixCreateTemplate(pixs);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);

    rotateAMLow(datad, w, h, wpld, datas, wpls, angle);

    return pixd;
}


/*------------------------------------------------------------------*
 *                    Rotation about the UL corner                  *
 *------------------------------------------------------------------*/
/*!
 *  pixRotateAMCorner()
 *
 *         Input: pixs
 *                angle (radians; clockwise is positive)
 *         Return:  pixd, or null on error
 *
 *         Note: this rotates the image about the UL corner.
 *               A positive angle gives a clockwise rotation.
 */
PIX *
pixRotateAMCorner(PIX     *pixs,
	          FLOAT32  angle)
{
INT32         w, h, wpls, wpld;
UINT32       *datas, *datad;
PIX    	     *pixd;

    PROCNAME("pixRotateAMCorner");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 8)
        return (PIX *)ERROR_PTR("pixs must be 8 bpp", procName, NULL);

    if (ABS(angle) < VERY_SMALL_ANGLE)
	return pixClone(pixs);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);
    pixd = pixCreateTemplate(pixs);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);

    rotateAMCornerLow(datad, w, h, wpld, datas, wpls, angle);

    return pixd;
}


