/*contains function to calculate, sort and evaluate the asymmetry score in each lane*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "asymmetry.h"
#include "position.h"
#include "checkresult.h"

/*sorting code based on code taken from http://linux.wku.edu/~lamonml/algor/sort/shell.html*/
void shellSort(float numbers[], int array_size)
{
  int i, j, increment;
  float temp;

  increment = 3;
  while (increment > 0)
  {
    for (i=0; i < array_size; i++)
    {
      j = i;
      temp = numbers[i];
      while ((j >= increment) && (numbers[j-increment] > temp))
      {
        numbers[j] = numbers[j - increment];
        j = j - increment;
      }
      numbers[j] = temp;
    }
    if (increment/2 != 0)
      increment = increment/2;
    else if (increment == 1)
      increment = 0;
    else
      increment = 1;
  }
}

/*finds the upper whisker in a set of numbers for identifying outliers by the whisker_factor x IQR criterion*/
static float findUpperOutlierCutoff(float *indata, int length, float whisker_factor)
{
  float median_pos, upper_quad_pos;
  int median_low, median_high, upper_quad_low, upper_quad_high;
  float median, upper_quad;
  
  median_pos = (float) (length-1) / 2.0;
  upper_quad_pos = (float) (length-1) * 3.0 / 4.0;

  median_low = (int) floor(median_pos);
  median_high = (int) ceil(median_pos); 
  upper_quad_low = (int) floor(upper_quad_pos);
  upper_quad_high = (int) ceil(upper_quad_pos); 

  shellSort(indata, length);
  
  if (median_low == median_high) {
    median = indata[median_low];
  } else {
    median = indata[median_high]*((float)median_high-median_pos) + indata[median_low]*(median_pos-(float)median_low);
  }
  if (upper_quad_low == upper_quad_high) {
    upper_quad = indata[upper_quad_low];
  } else {
    upper_quad = indata[upper_quad_high]*((float)upper_quad_high-upper_quad_pos) + indata[upper_quad_low]*(upper_quad_pos-(float)upper_quad_low);
  }

  return upper_quad+(upper_quad-median)*whisker_factor;
}


/*calculate min asymmetry value in a given vector for all possible 
*returns -1 if windowsize too large
*/
static float asymmetry(float *invalues, int length, int windowsize, int maxoffset,  int diffcap, float *offsetused)
{
  float asym=0;
  float curasym;
  float refcenter;
  int index;
  float halfwidth, halflane;
  float valueleft;
  float valueright;
  
  halfwidth = (float)windowsize/2.0;
  halflane = (float)length/2.0;
  
  if(ceil(halflane-maxoffset-halfwidth)<0 || ceil(halflane+maxoffset+halfwidth>length))
    {
      fprintf(stderr, "ERROR:  asymetry window and offset size not valid\n");
      exit(1);
    }
  *offsetused = 0;  
  for(refcenter = (float)-maxoffset; refcenter<=(float)maxoffset;refcenter+= 0.5)
    {
      curasym=0;
      for(index=1; index<=(int)halfwidth; index++)
	{

	  valueright=invalues[(int)floor(halflane+refcenter-index+0.25)];/*add in .25 to avoid imprecision*/
	  
	  if (ceil(halflane+refcenter)==floor(halflane+refcenter)) valueleft = invalues[(int)floor(halflane+refcenter+index-1+0.25)];
	  else valueleft = invalues[(int)floor(halflane+refcenter+index+0.25)];
	  /*should not contribute to asym if difference too small, avoid low mean value from inflating asym value*/
	  if (fabs(valueleft-valueright)>diffcap) curasym+=(valueleft-valueright)/(valueleft+valueright);
	}
      if(((fabs(curasym)<fabs(asym))||asym==0)&&(curasym!=0))
	{
	 *offsetused = refcenter;
	  asym=curasym;
	}
    }
  return fabs(asym)/(float) (2*halfwidth);
}

/*based on Peak and Node Positions, the area of a given lane is cropped out, and from this the median row is extracted and returned*/
void followLane(PIX *pixGel, float **PeakPosition, float *NodePosition, struct GelInfo *GelInfo, int laneID, float **laneintensity)
{

int nodeID;
int xindex, xstop, ylow, yhigh, yindex;
float y, slope;
int	halfwidth, output_width;
UINT32  *linelow, *linehigh, *datas;
INT32   wpls;

 halfwidth = ((int)(GelInfo->LaneSpacing + 0.5)) / 2;  
 output_width = 2 * halfwidth + 1;

 datas = pixGetData(pixGel);
 wpls = pixGetWpl(pixGel);

 for (yindex = halfwidth; yindex >= -halfwidth; yindex--) {
   for (nodeID = 0; nodeID < GelInfo->NumberOfZones+1; nodeID++) {

   slope = (fabs(PeakPosition[nodeID+1][laneID])-fabs(PeakPosition[nodeID][laneID]))/(NodePosition[nodeID+1]-NodePosition[nodeID]);
     xindex = (int)(NodePosition[nodeID]+0.5);
     xstop = (int)(NodePosition[nodeID+1]+0.5);
     if (nodeID == GelInfo->NumberOfZones) xstop++;
     while (xindex < xstop) {
       y = slope * ((float)xindex-NodePosition[nodeID])+fabs(PeakPosition[nodeID][laneID])-0.5;
       ylow=floor(y);
       yhigh=ceil(y);
       linelow = datas + (ylow+yindex) * wpls;
       if (ylow == yhigh) {
	 laneintensity[xindex][yindex+halfwidth] = GET_DATA_BYTE(linelow, xindex);
       } else {
	 linehigh = datas + (yhigh+yindex) * wpls;
	 laneintensity[xindex][yindex+halfwidth]= (float)GET_DATA_BYTE(linelow, xindex)*((float)yhigh-y) +
	   (float)GET_DATA_BYTE(linehigh, xindex)*(y-(float)ylow);
       }
     xindex++;
     }
   }
 }
}

/*check symmetry of all the lanes, in groups of zones.  A corrected image of each lane is extracted using peak and node positions. 
 * The asymmetry is calculated as the difference of intensity the same distance from a center.  The center is shifted up and down,
 * and the lowest value of the asymmetry is used. Only columns with sufficient content is included.  
 * The mean of the asym score and the offset used in each segment is reported. 
*/
int checkSymmetry(FILE *ERROROUT_FILE, FILE *FLAGOUT_FILE, PIX *pixGel, int* ZoneID, float **PeakPosition, float *NodePosition, struct GelInfo *GelInfo, struct AnaCutoff *AnaCutoff, float **Asym, float **Laneoffset, int *con_asym_lanes,int Verbose)
{

int laneID, badlaneID;
float **laneintensity;
float *colrange, *meanrange;
int *zonesize;
int   halfwidth, output_width;
int col, row, index, zoneindex;
int windowsize;
float maxint, minint;
int *colincluded;
int nasym=0;
int i, cur_con_asym_lanes;
float asymcutoff;
float *buffer;
int maxoffset;
float offsetused;
 /*to add to anacutoff*/
 int diffcap = (int) floor(AnaCutoff->asym_threshold*(pow(2,pixGetDepth(pixGel))-1));

 halfwidth = ((int)(GelInfo->LaneSpacing + 0.5)) / 2;  
 output_width = 2 * halfwidth + 1;
 windowsize = (int) ((float)output_width*AnaCutoff->asym_window_factor+0.5);
 maxoffset =(int) ((AnaCutoff->asym_max_offset*output_width)+.5);

 laneintensity=(float**) malloc((GelInfo->ImageWidth) * sizeof(float*));
 for (col = 0; col < GelInfo->ImageWidth; col++) 
   laneintensity[col] =(float*)  calloc(output_width, sizeof(float));
 colrange = (float*) calloc(GelInfo->ImageWidth, sizeof(float));
 meanrange = (float*) calloc(GelInfo->NumberOfZones, sizeof(float));
 zonesize = (int*) calloc(GelInfo->NumberOfZones, sizeof(int));
 colincluded = (int*) calloc(GelInfo->NumberOfZones, sizeof(int));



 for(col = GelInfo->StartPreData; col<GelInfo->StopData; col++)
   {
     zonesize[ZoneID[col]]++;
   }

 for(laneID=0; laneID<GelInfo->NumberOfLanes; laneID++)
   {


     followLane(pixGel, PeakPosition, NodePosition, GelInfo, laneID, laneintensity);
   
     /*reset meanrange and colincluded to 0 for new lane*/
     for(zoneindex = 0; zoneindex<GelInfo->NumberOfZones; zoneindex++)
       {
	 meanrange[zoneindex]=0;
	 colincluded[zoneindex]=0;
       }


     /*find the mean intensity range for each column to decide if the column needs to be included in the asymmetry*/
     for(col = GelInfo->StartPreData; col<GelInfo->StopData; col++)
       {
	 maxint = laneintensity[col][0];
	 minint = laneintensity[col][0];
	 for(row=0;row<output_width;row++)
	   {
	     if (maxint<laneintensity[col][row])maxint = laneintensity[col][row];
	     if (minint>laneintensity[col][row])minint = laneintensity[col][row];
	   } 
	 colrange[col]=maxint-minint;
	 meanrange[ZoneID[col]]+=colrange[col];
       }
     for(zoneindex = 0; zoneindex<GelInfo->NumberOfZones; zoneindex++) meanrange[zoneindex]/=(float)zonesize[zoneindex];

     /*include in symmetry score of zone if the range is greater than the mean range in the zone*/
     for(col = GelInfo->StartPreData; col<GelInfo->StopData; col++)
       {
	 if(colrange[col]>meanrange[ZoneID[col]]) /*enough range*/
	   {
	     Asym[ZoneID[col]][laneID]+= asymmetry(laneintensity[col], output_width, windowsize, maxoffset, diffcap, &offsetused);
 	     Laneoffset[ZoneID[col]][laneID]+=offsetused;
/* 	    if (offsetused<0) fprintf(stdout, "%i ", offsetused); */
	     colincluded[ZoneID[col]]++;
	   }
       }
     /*normalize by the number of columns included, if 0 set asym to 0*/
     for(zoneindex = 0; zoneindex<GelInfo->NumberOfZones; zoneindex++)
       {
	 if(colincluded[zoneindex]==0) 
	   {
	     Asym[zoneindex][laneID]=0;
	     Laneoffset[zoneindex][laneID]=0;
	   }
	 else  
	   {
	     Asym[zoneindex][laneID]/=(float)colincluded[zoneindex];
	     Laneoffset[zoneindex][laneID]/=(float)colincluded[zoneindex];
	   }
       }
   }

 buffer = (float *)calloc(GelInfo->NumberOfLanes, sizeof(float)); 
 /*check for consec asym marker and set all asym for peaks in offending group to negative*/
 for(zoneindex = 0; zoneindex < GelInfo->NumberOfZones; zoneindex++)
   {
     con_asym_lanes[zoneindex] = 0; 
     cur_con_asym_lanes =0;
     for (laneID = 0; laneID < GelInfo->NumberOfLanes; laneID++)
       {
	 buffer[laneID] = Asym[zoneindex][laneID];
       }
     asymcutoff = findUpperOutlierCutoff(buffer, GelInfo->NumberOfLanes, AnaCutoff->asym_whisker_factor);

     for (laneID = 0; laneID < GelInfo->NumberOfLanes; laneID++)
       {
	 if (Asym[zoneindex][laneID]>asymcutoff)
      	   {
	     cur_con_asym_lanes++;
	   }
	 else
	   {
	     if(cur_con_asym_lanes>con_asym_lanes[zoneindex]) con_asym_lanes[zoneindex]=cur_con_asym_lanes;
	     if(cur_con_asym_lanes>= AnaCutoff->consec_asym_soft_cutoff)
	       {
		 for(badlaneID = laneID-1; badlaneID >= laneID-cur_con_asym_lanes;badlaneID--) /*setting all lanes in the preceeding set to negative*/
		   {
		     Asym[zoneindex][badlaneID]*=-1;
		   }
	       }
	     cur_con_asym_lanes=0; /*restart count*/
	   }
       }


    /*check interpolated marker and interpolated data if necessary to fail gel*/
     if(con_asym_lanes[zoneindex] >= AnaCutoff->consec_asym_hard_cutoff)
       {
	 GelInfo->GelPassed = 0;
	 fprintf(ERROROUT_FILE, "Found %i consecutive asymmmetric lanes in zone %i; lanes:", con_asym_lanes[zoneindex], zoneindex+1);	 
	 if (Verbose)  fprintf(stderr, "Found %i consecutive asymmmetric lanes in zone %i; lanes:", con_asym_lanes[zoneindex], zoneindex+1);
	 
	 for (laneID = 0; laneID < GelInfo->NumberOfLanes; laneID++)
	   {   
	     if (fabs(Asym[zoneindex][laneID])>asymcutoff)
	       {
		 if (Asym[zoneindex][laneID]<0) /*output all offending lanes*/
		   {
		     if(laneID==0 || Asym[zoneindex][laneID-1]<0) 
		       {
			 if (Verbose) fprintf(stderr, " %i", GelInfo->NumberOfLanes-1 - laneID);
			 fprintf(ERROROUT_FILE, " %i", GelInfo->NumberOfLanes-1 - laneID);
		     }
		     else /*add comma for new group*/
		       {
			 if (Verbose) fprintf(stderr, ", %i", GelInfo->NumberOfLanes-1 - laneID);
			 fprintf(ERROROUT_FILE, ", %i", GelInfo->NumberOfLanes-1 - laneID);
			 
		       }
		   }
	       }
	   }
	 if(Verbose) fprintf(stderr, ". Gel FAILED\n");
	 fprintf(ERROROUT_FILE, ". Gel FAILED\n");     
       }
   }

 
 /*output to file before correction*/
 fprintf(FLAGOUT_FILE, "Asymmetry for each lane in each zone before adjustment:\n"); 
 for (laneID = GelInfo->NumberOfLanes-1; laneID >=0; laneID--)
   {
     for(zoneindex=0; zoneindex<GelInfo->NumberOfZones; zoneindex++)
       {
	 fprintf(FLAGOUT_FILE, "\t%f", Asym[zoneindex][laneID]);
       }
     fprintf(FLAGOUT_FILE, "\n");
   }
 fprintf(FLAGOUT_FILE, "Lane offset for each lane in each zone before adjustment:\n"); 
 for (laneID = GelInfo->NumberOfLanes-1; laneID >=0; laneID--)
   {
     for(zoneindex=0; zoneindex<GelInfo->NumberOfZones; zoneindex++)
       {
	 fprintf(FLAGOUT_FILE, "\t%f", Laneoffset[zoneindex][laneID]);
       }
     fprintf(FLAGOUT_FILE, "\n");
   }

  
 for (index = 0; index < GelInfo->ImageWidth; index++) {
   free(laneintensity[index]);
 }
 free(buffer);
 free(laneintensity);
 free(colrange);
 free(zonesize);
 free(meanrange);
 free(colincluded);
  return nasym;

}

/*correct all peaks in predata identified to have low symmetry (noisy area) with filtered peak projection*/
int correctAsymmetry(INT32 **ZoneProjection, float **PeakPosition, float *Spacing, struct GelInfo *GelInfo, float **Asym, int MarkerInc)
{

  int ncorrected =0;
  int needfilling, laneID, nodeID, zoneID, i;
  float tolerance = fabs(GelInfo->LaneSpacing) / TOLERANCEFACTOR;  

  /*correcting peaks*/
  for(laneID = 0; laneID<GelInfo->NumberOfLanes;laneID++)
    {
      for(zoneID = 0; zoneID<GelInfo->NumberOfZones; zoneID++)
	{
	  nodeID=zoneID+1;
	  needfilling =0;
	  if(Asym[zoneID][laneID] < 0) /*need adjusting*/
	    {
	      /* fprintf(stderr, "correcting lane %i zone %i\n", GelInfo->NumberOfLanes-1-laneID, zoneID+1); */
	      
	      /*still respect marker dominance*/
	      if(MarkerInc==0|| fabs(Spacing[zoneID]-GelInfo->LaneSpacing) < tolerance || (laneID-GelInfo->MarkerOffset)%GelInfo->MarkerPeriod ==0)
		{
		  ncorrected++;
		  /*correct the one peak, will be 0 if not matched*/
		  matchSinglePeak(ZoneProjection, GelInfo, PeakPosition, zoneID, laneID);
		  if ( PeakPosition[nodeID][laneID]==0) needfilling=1;
		  /*all estimated peaks based on outlier has to be refilled*/
		  i = laneID - 1;
		  while (i >= 0) {
		    if (PeakPosition[nodeID][i] < 0) {
		      PeakPosition[nodeID][i] = 0;
		      i--;
		      needfilling=1;
		    } else {
		      break;
		    }
		  }
		  i = laneID + 1;
		  while (i < GelInfo->NumberOfLanes) {
		    if (PeakPosition[nodeID][i] < 0) {
		      PeakPosition[nodeID][i] = 0;
		      i++;
		      needfilling=1;
		    } else {
		      break;
		    }
		  }
		  if (needfilling) fillEmpties(PeakPosition[nodeID], GelInfo);	 
		}
	    }
	}
    }
  
  return ncorrected;
}

