#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "spacing.h"

  /*finds the maximum of FFT power spectrum in  y-direction  within each zone, and using the result to get the dominant period between marker lanes*/
static void fillSpacing(INT32 **ZoneProjection, struct GelInfo *GelInfo, float *Spacing) {
fftw_complex *fft;
double *spectrum, *power;
fftw_plan plan;
int length, Nd2p1, specID, channel, chMax, StartSearch;
double Max, peakposition;
  
  length = GelInfo->ImageHeight;
  Nd2p1 = length/2 + 1;
  spectrum = fftw_malloc(sizeof(double) * length);
  fft = fftw_malloc(sizeof(fftw_complex) * Nd2p1);
  power = calloc(Nd2p1, sizeof(double));
  plan = fftw_plan_dft_r2c_1d(length, spectrum, fft, FFTW_ESTIMATE);

  StartSearch = GelInfo->NumberOfLanes / GelInfo->MarkerPeriod; /*initial guess, max number of marker lanes, avoids DC term*/

  for (specID = 0; specID < GelInfo->NumberOfZones; specID++) {
    for (channel = 0; channel < length; channel++) {
      spectrum[channel] = (double) (ZoneProjection[specID][channel]); /*YProjection of a zone*/
    }
    fftw_execute(plan);
    for (channel = 0; channel < Nd2p1; channel++) {
      power[channel] = (fft[channel][0]*fft[channel][0])+(fft[channel][1]*fft[channel][1]);
    }
    chMax = 1;
    Max = 0;
    for (channel = StartSearch; channel < Nd2p1-1; channel++) {
        if(power[channel] > Max) {
          chMax = channel;
	  Max = power[channel];
        }
    }
    /*weighted average of peak and neighbours to give better noise rejection, and higher resolution*/
    peakposition = (power[chMax-1]*(float)(chMax-1)) + (power[chMax]*(float)chMax) + (power[chMax+1]*(float)(chMax+1));
    peakposition /= power[chMax-1] + power[chMax] + power[chMax+1];
    Spacing[specID] = (float)length / peakposition; /*height/dominant frequency gives most likely spacing in pixels, this can be markerspacing or lane spacing*/
  }
  
  fftw_destroy_plan(plan);
  fftw_free(spectrum);
  fftw_free(fft);
  free(power);
}

/*find lane spacing, defined as the minimum in all the zones 
 *that is seen in more than one zone
 *returns -1 if no valid spacing found
 */
float findSpacing (FILE *FLAGOUT_FILE, INT32 **ZoneProjection, struct GelInfo *GelInfo, float *Spacing,
                  int Verbose) {

int i, j, seenNtimes;
int *ISpacing, imin, nmin;
float minimum, GenericLaneSpacing;

  fillSpacing(ZoneProjection, GelInfo, Spacing);
  if (Verbose) fprintf(stderr, "Dominant spacing for each zone:\n");
  
  fprintf(FLAGOUT_FILE, "Dominant spacing:");
  for (i = 0; i < GelInfo->NumberOfZones; i++) {
      if(Verbose) fprintf(stderr, "\tzone %d = %.2f\n", i, Spacing[i]);
      fprintf(FLAGOUT_FILE, "\t%.2f", Spacing[i]);
    }     
  fprintf(FLAGOUT_FILE, "\n");
    
  ISpacing = (int *)calloc(GelInfo->NumberOfZones, sizeof(int));
  for (i = 0; i < GelInfo->NumberOfZones; i++) {
    ISpacing[i] = (int) (Spacing[i] + 0.5);
  }
  for (i = 0; i < GelInfo->NumberOfZones; i++) {
    seenNtimes = 0;
    for (j = 0; j < GelInfo->NumberOfZones; j++) {
      if (ISpacing[i] == ISpacing[j]) {
        seenNtimes++;
	if (seenNtimes > 1) break;
      }
    }
    if (seenNtimes < 2) ISpacing[i] = GelInfo->ImageHeight;
  } /*there must be at least one other zone with same ISpacing for the result to be used, spacing is set to image height (invalid) otherwise*/
  imin = GelInfo->ImageHeight;
  for (i = 0; i < GelInfo->NumberOfZones; i++) {
    if (ISpacing[i] < imin) imin = ISpacing[i];
  }
  if (imin < GelInfo->ImageHeight) {
    nmin = 0;
    GenericLaneSpacing = 0.0;
    for (i = 0; i < GelInfo->NumberOfZones; i++) {
      if (ISpacing[i] == imin) {
        nmin++;
        GenericLaneSpacing += Spacing[i];
      }
    }
    if (nmin > 1) {
      GenericLaneSpacing /= (float)nmin; /*average of Spacings with same minimum ISpacing*/
      if (GenericLaneSpacing * (GelInfo->NumberOfLanes-1) >= GelInfo->ImageHeight) {
/*check if we did not find lane spacing or we assume marker spacing found*/
        if (Verbose) 
          fprintf(stderr, "Spacing of %d is too large to fit %d lanes, will assume we found the marker spacing\n",imin,GelInfo->NumberOfLanes);
        GenericLaneSpacing /= (float)GelInfo->MarkerPeriod;
      } 
      if (Verbose) {
        fprintf(stderr,"Generic lane spacing = %.2f\n", GenericLaneSpacing);
      }
      fprintf(FLAGOUT_FILE,"Generic lane spacing:\t%.2f\n", GenericLaneSpacing);
    }
  } else {
    GenericLaneSpacing = -1.0; /*invalid spacing*/
  }
  free(ISpacing);
  return GenericLaneSpacing;
}
